"""auxilary functions for PET list-mode data processing."""

__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"

#-------------------------------------------------------------------------------

import numpy as np
from math import pi
import sys
import os
pjoin = os.path.join
import glob

import scipy.ndimage as ndi
from pkg_resources import resource_filename

import nibabel as nib
import pydicom as dcm
import re

from niftypet import nimpa

import resources
import mmr_auxe

#=================================================================================================
def create_dir(pth):
    if not os.path.exists(pth):    
        os.makedirs(pth)

def fwhm2sig(fwhm):
    Cnt = resources.get_mmr_constants()
    return (fwhm/Cnt['SZ_VOXY']) / (2*(2*np.log(2))**.5)

#=================================================================================================
def lm_pos(datain, Cnt):
    '''get the position of table and gantry offset from the DICOM list-mode file'''

    if 'lm_dcm' in datain and os.path.isfile(datain['lm_dcm']):
        dhdr = dcm.read_file(datain['lm_dcm'])
    elif 'lm_ima' in datain and os.path.isfile(datain['lm_ima']):
        dhdr = dcm.read_file(datain['lm_ima'])
    else:
        print 'e> DICOM list-mode data not found!'
        return None

    #---find GantryOffset
    if dhdr[0x0018, 0x1020].value == 'syngo MR B20P'  or dhdr[0x0018, 0x1020].value == 'syngo MR E11':
        flip = 1
        if [0x29,0x1120] in dhdr:
            csainfo = dhdr[0x29,0x1120].value
        else:
            print 'e> DICOM field [0x29,0x1120] not found!'
            return None
    # this is for older scanner software
    elif dhdr[0x0018, 0x1020].value == 'syngo MR B18P':
        flip = -1
        if [0x29,0x1020] in dhdr:
            csainfo = dhdr[0x29,0x1020].value
        elif [0x29,0x1120] in dhdr:
            csainfo = dhdr[0x29,0x1120].value
        else:
            print 'e> DICOM field [0x29,0x1020] not found!'
            return None
    else:
        print 'e> unknown scanner software version!'
        sys.exit()

    fi = re.search(r'GantryOffset(?!_)', csainfo).start() #csainfo.find('GantryOffset')
    #regular expression for the needed three numbers
    p = re.compile(r'-?\d.\d{4,10}')
    xyz = p.findall(csainfo[fi:fi+200])
    #offset in cm
    xoff = float(xyz[0])/10
    yoff = float(xyz[1])/10
    zoff = float(xyz[2])/10
    goff = flip * np.array([xoff, yoff, zoff])
    if Cnt['VERBOSE']: print 'i> gantry offset:', goff

    fi = csainfo.find('TablePositionOrigin')
    #regular expression for the needed three numbers
    tpostr = csainfo[fi:fi+200]
    tpo = re.sub(r'[^a-zA-Z0-9\-]', '', tpostr).split('M')
    tpozyx = np.array([float(tpo[-1]), float(tpo[-2]), float(tpo[-3])])
    if Cnt['VERBOSE']: print 'i> table position origin:', tpozyx

    return goff, tpozyx

#=================================================================================================
def hdr_lm(datain, Cnt):
    '''Get the headers from DICOM list-mode data file'''
    if 'lm_dcm' in datain and os.path.isfile(datain['lm_dcm']):
        dhdr = dcm.read_file(datain['lm_dcm'])
    elif 'lm_ima' in datain and os.path.isfile(datain['lm_ima']):
        dhdr = dcm.read_file(datain['lm_ima'])
    else:
        print 'e> DICOM list-mode data not found!'
        return None

    # list possible DICOM locations for list-mode interfile header 
    lmhdr_locations = [[0x29,0x1010], [0x29,0x1110]]

    # for newer scanner software
    if dhdr[0x0018, 0x1020].value == 'syngo MR B20P' or dhdr[0x0018, 0x1020].value == 'syngo MR E11':
        # interfile header
        if [0x29,0x1010] in dhdr:
            lmhdr = dhdr[0x29,0x1010].value
            if Cnt['VERBOSE']: print 'i> got LM interfile.'
        else:
            print 'w> DICOM field [0x29,0x1010] not found!'
            lmhdr = None

        #CSA Series Header Info
        if [0x29,0x1120] in dhdr:
            csahdr = dhdr[0x29,0x1120].value
            if Cnt['VERBOSE']: print 'i> got CSA info.'
        else:
            print 'e> DICOM field [0x29,0x1120] not found!'
            return lmhdr, None
    
    # for older scanner software
    elif dhdr[0x0018, 0x1020].value == 'syngo MR B18P':
        
        # find interfile header
        found_lmhdr = False
        for loc in lmhdr_locations:
            if loc in dhdr:
                lmhdr = dhdr[loc].value
                if '!INTERFILE' in lmhdr and 'start horizontal bed position' in lmhdr:
                    if Cnt['VERBOSE']: print 'i> got LM interfile from [', hex(loc[0]),',', hex(loc[1]), ']'
                    found_lmhdr = True
                    break
        if not found_lmhdr:            
            print 'w> DICOM field with LM interfile header has not been found!'
            lmhdr = None
        
        #CSA Series Header Info
        if [0x29,0x1020] in dhdr:
            csahdr = dhdr[0x29,0x1020].value
            if Cnt['VERBOSE']: print 'i> got CSA info.'
        elif [0x29,0x1120] in dhdr:
            csahdr = dhdr[0x29,0x1120].value
            if Cnt['VERBOSE']: print 'i> got CSA info (may not be accurate, please check).'
        else:
            print 'e> DICOM field [0x29,0x1020] not found!'
            return lmhdr, None

    return lmhdr, csahdr

def vh_bedpos(datain, Cnt):
    ihdr, csainfo = hdr_lm(datain, Cnt)

    #start horizontal bed position
    p = re.compile(r'start horizontal bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    hbedpos = 0.1*float(ihdr[m.start()+fi+1:m.end()])

    #start vertical bed position
    p = re.compile(r'start vertical bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    vbedpos = 0.1*float(ihdr[m.start()+fi+1:m.end()])

    return vbedpos, hbedpos

def hmu_resample0(hmupos, parts ,Cnt):
    
    #output image sampling
    Cim = {
        'VXSRx':Cnt['SO_VXX'],
        'VXSRy':Cnt['SO_VXY'],
        'VXSRz':Cnt['SO_VXZ'],
        'VXNRx':Cnt['SO_IMX'],
        'VXNRy':Cnt['SO_IMY'],
        'VXNRz':Cnt['SO_IMZ']
    }
    #voxel position/offset
    Cim['OFFRx'] = -0.5*Cim['VXNRx']*Cim['VXSRx']#-0.5*Cim['VXSRx']
    Cim['OFFRy'] = -0.5*Cim['VXNRy']*Cim['VXSRy']#-0.5*Cim['VXSRy']
    Cim['OFFRz'] = -0.5*Cim['VXNRz']*Cim['VXSRz']-hmupos[0]['HBedPos']

    Trnsl = (0.0, 0.0, 0.0)
    #transformation matrix
    A = np.array(
        [[ 1., 0., 0.,  Trnsl[0] ],
        [  0., 1., 0.,  Trnsl[1] ],
        [  0., 0., 1.,  Trnsl[2] ],
        [  0., 0., 0.,  1. ]], dtype=np.float32
        )

    imr = np.zeros( (Cnt['SO_IMZ'],Cnt['SO_IMY'],Cnt['SO_IMX']), dtype=np.float32)
    #===== Go through the hardware mu-map parts =====
    for i in parts:
        Cim['VXSOx'] = hmupos[i]['ivs'][2]
        Cim['VXSOy'] = hmupos[i]['ivs'][1]
        Cim['VXSOz'] = hmupos[i]['ivs'][0]
        Cim['VXNOx'] = hmupos[i]['shape'][2]
        Cim['VXNOy'] = hmupos[i]['shape'][1]
        Cim['VXNOz'] = hmupos[i]['shape'][0]
        #original image offset
        Cim['OFFOx'] = hmupos[i]['vpos'][2]
        Cim['OFFOy'] = hmupos[i]['vpos'][1]
        Cim['OFFOz'] =-hmupos[i]['vpos'][0]

        #resample!
        if i==4:
            #does the bed just partly (no point doing all the long bed)
            offresZ = (-.5*Cnt['SO_IMZ']*Cnt['SO_VXZ']-hmupos[0]['HBedPos'])
            #excess of the hrdwr mu-map axially
            excemuZ = offresZ - (-hmupos[4]['vpos'][0])
            excevox = int( excemuZ/hmupos[4]['ivs'][0] ) - 5# with extra margin of 5
            newoffZ = -hmupos[4]['vpos'][0] + excevox*hmupos[4]['ivs'][0]
            #number of voxels included axially
            inclvox = Cnt['SO_IMZ']*Cnt['SO_VXZ']/hmupos[4]['ivs'][0] + 10 #with extra margin...
            #truncate the image
            im = hmupos[i]['img'][excevox:excevox+inclvox,:,:]
            #update dictionary Cim
            Cim['OFFOz'] = newoffZ
            Cim['VXNOz'] = im.shape[0]
            imr += nimpa.prc.improc.resample(im, A, Cim)

        else:
            imr += improc.resample(hmupos[i]['img'], A, Cim)

    return imr
    
    


#=================================================================================================
def time_diff_norm_acq(datain):

    if 'lm_dcm' in datain and os.path.isfile(datain['lm_dcm']):
        l = dcm.read_file(datain['lm_dcm'])
    elif 'lm_ima' in datain and os.path.isfile(datain['lm_ima']):
        l = dcm.read_file(datain['lm_ima'])
    else:
        print 'e> dicom header of list-mode data does not exist.'
        return None
    
    # acq date
    s = l[0x08,0x21].value
    y = int(s[:4])
    m = int(s[4:6])
    d = int(s[6:8])
    # acq time
    s = l[0x08,0x32].value
    hrs = int(s[:2])
    mns = int(s[2:4])
    sec = int(s[4:6])

    # calib date
    s = l[0x18,0x1200].value
    cy = int(s[:4])
    cm = int(s[4:6])
    cd = int(s[6:8])
    # calib time
    s = l[0x18,0x1201].value
    chrs = int(s[:2])
    cmns = int(s[2:4])
    csec = int(s[4:6])

    tdiff = (hrs*3600 + mns*60 + sec) - (chrs*3600 + cmns*60 + csec)
    dhrs = tdiff/3600
    dmns = (tdiff - 3600*dhrs)/60
    if dhrs>12:
        print 'w> time difference between calibration and acquisition is:', dhrs,'hrs and',dmns,'mins'

    if np.sum([cy-y, cm-m, cd-d])!=0:
        print 'w> daily QC/calibration was performed on different day !!!'
        print [cy, cm, cd], 'vs', [y,m,d]


def timings_from_list(flist, offset=0):
    '''
    Get start and end frame timings from a list of dynamic PET frame definitions.
    flist can be 1D list of time duration for each dynamic frame, e.g.: flist = [15, 15, 15, 15, 30, 30, 30, ...]
    or a 2D list of lists having 2 entries: first for the number of repetitions and the other for the frame duration,
    e.g.: flist = [[4,15], [3,15], ...].
    offset adjusts for the start time (usually when prompts are strong enough over randoms)
    The output is a dictionary:
    out['timings'] = [[0, 15], [15, 30], [30, 45], [45, 60], [60, 90], [90, 120], [120, 150], ...]
    out['total'] = total time
    out['frames'] = array([ 15,  15,  15,  15,  30,  30,  30,  30, ...])

    '''
    if not isinstance(flist, list):
        raise TypeError('Wrong type of frame data input')
    if all([isinstance(t,(int, np.int32, np.int16, np.int8, np.uint8, np.uint16, np.uint32)) for t in flist]):
        tsum = offset
        # list of frame timings
        if offset>0:
            t_frames = [[0, offset]]
        else:
            t_frames = []
        for i in range(len(flist)):
            # frame start time
            t0 = tsum
            tsum += flist[i]
            # frame end time
            t1 = tsum
            # append the timings to the list
            t_frames.append([t0, t1])
        frms = np.uint16(flist)
    elif all([isinstance(t,list) and len(t)==2 for t in flist]):
        if offset>0:
            flist.insert(0,[1,offset])
            farray = np.asarray(flist, dtype=np.uint16)
        else:
            farray = np.array(flist)
        # number of dynamic frames
        nfrm = np.sum(farray[:,0])
        # list of frame duration
        frms = np.zeros(nfrm,dtype=np.uint16)
        #frame iterator
        fi = 0
        #time sum of frames
        tsum = 0
        # list of frame timings
        t_frames = []
        for i in range(0, farray.shape[0]):
            for t in range(0, farray[i,0]):
                # frame start time
                t0 = tsum
                tsum += farray[i,1]
                # frame end time
                t1 = tsum
                # append the timings to the list
                t_frames.append([t0, t1])
                frms[fi] = farray[i,1]
                fi += 1
    else:
        raise TypeError('Unrecognised data input.')
    # prepare the output dictionary
    out = {'total':tsum, 'frames':frms, 'timings':t_frames}
    return out
#=================================================================================================




    
#=================================================================================================
def axial_lut(Cnt):
    ''' Creates lookup tables (LUT) for linear indexes along the diagonals of Michelogram
    for span-11 calculations done on GPU.
    '''

    NRNG = Cnt['NRNG']

    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        NSN1_c = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction) 
        if NRNG_c==64:
            NSN1_c -= 12
        SEG0_c = 2*NRNG_c-1
    else:
        NRNG_c = NRNG
        NSN1_c = Cnt['NSN1']
        if Cnt['RNG_END']!=NRNG or Cnt['RNG_STRT']!=0:
            print 'e> the reduced axial FOV only works in span-1!'
            return None

    
    #ring dimensions
    rng = np.zeros((NRNG,2), dtype = np.float32)
    z = -.5*NRNG*Cnt['AXR']
    for i in range(NRNG):
        rng[i,0] = z
        z += Cnt['AXR']
        rng[i,1] = z

    #--create mapping from ring difference to segment number
    #ring difference range
    rd = range(-Cnt['MRD'],Cnt['MRD']+1)
    #ring difference to segment
    rd2sg = -1*np.ones((len(rd),2,), dtype=np.int32)
    for i in range(len(rd)):
        for iseg in range(len(Cnt['MNRD'])):
            if ( rd[i]>=Cnt['MNRD'][iseg] ) and ( rd[i]<=Cnt['MXRD'][iseg] ):
                rd2sg[i,:] = np.array([rd[i], iseg])
    #--

    #create two Michelograms for segments (Mseg) 
    #and absolute axial position for individual sinos (Mssrb) which is single slice rebinning
    Mssrb = -1*np.ones((NRNG,NRNG), dtype=np.int32)
    Mseg = -1*np.ones((NRNG,NRNG), dtype=np.int32)
    for r1 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):
        for r0 in range(Cnt['RNG_STRT'], Cnt['RNG_END']):
            if abs(r1-r0)>Cnt['MRD']:
                continue
            ssp = r0+r1  #segment sino position (axially: 0-126)
            rd = r1-r0
            jseg = rd2sg[rd2sg[:,0]==rd, 1]
            Mssrb[r1,r0] = ssp 
            Mseg[r1,r0] = jseg #negative segments are on top diagonals

    # np.savetxt("Mssrb.csv", Mssrb, delimiter=",", fmt='%d')
    # np.savetxt("Mseg.csv", Mseg, delimiter=",", fmt='%d')


    #create a Michelogram map from rings to sino number in span-11 (1..837)
    Msn = -1*np.ones((NRNG,NRNG), dtype=np.int32)
    #number of span-1 sinos per sino in span-11
    Mnos = -1*np.ones((NRNG,NRNG), dtype=np.int32)
    i = 0
    for iseg in range(0,len(Cnt['SEG'])):
        msk = (Mseg==iseg)
        Mtmp = np.copy(Mssrb)
        Mtmp[~msk] = -1
        uq = np.unique(Mtmp[msk])
        for u in range(0,len(uq)):
            #print i
            Msn [ Mtmp==uq[u] ] = i
            Mnos[ Mtmp==uq[u] ] = np.sum(Mtmp==uq[u])
            i += 1
    # np.savetxt("Mnos.csv", Mnos, delimiter=",", fmt='%d')
    # np.savetxt("Msn.csv", Msn, delimiter=",", fmt='%d')
    

    #====full LUT
    sn1_rno = np.zeros((NSN1_c,2), dtype=np.int16)
    sn1_ssrb= np.zeros((NSN1_c), dtype=np.int16)
    sn1_sn11= np.zeros((NSN1_c), dtype=np.int16)
    sn1_sn11no = np.zeros((NSN1_c), dtype=np.int8)
    sni = 0 #full linear index, upto 4084
    Msn1 = -1*np.ones((NRNG,NRNG), dtype=np.int16) #michelogram of sino numbers for spn-1
    for ro in range(0,NRNG):
        if ro==0:
            oblique = 1
        else:
            oblique = 2
        for m in range(oblique):
            strt = NRNG*(ro+Cnt['RNG_STRT']) + Cnt['RNG_STRT']
            stop = (Cnt['RNG_STRT']+NRNG_c)*NRNG
            step = NRNG+1
            for li in range(strt, stop, step): #goes along a diagonal started in the first row at r1
                #linear indecies of michelogram --> subscript indecies for positive and negative RDs
                if m==0:
                    r1 = li/NRNG
                    r0 = li - r1*NRNG
                else: #for positive now (? or vice versa)
                    r0 = li/NRNG
                    r1 = li - r0*NRNG
                #avoid case when RD>MRD
                if (Msn[r1,r0])<0:
                    continue

                sn1_rno[sni,0] = r0
                sn1_rno[sni,1] = r1

                sn1_ssrb[sni] = Mssrb[r1,r0]
                sn1_sn11[sni] = Msn[r0,r1]

                sn1_sn11no[sni] = Mnos[r0,r1]

                Msn1[r0,r1] = sni
                #--
                sni += 1

    #span-11 sino to SSRB
    sn11_ssrb = np.zeros(Cnt['NSN11'], dtype=np.int32); 
    sn11_ssrb[:] -= 1
    sn1_ssrno = np.zeros(Cnt['NSEG0'], dtype=np.int8)
    for i in range(NSN1_c):
        sn11_ssrb[sn1_sn11[i]] = sn1_ssrb[i]
        sn1_ssrno[sn1_ssrb[i]] += 1

    sn11_ssrno = np.zeros(Cnt['NSEG0'], dtype=np.int8)
    for i in range(Cnt['NSN11']):
        if sn11_ssrb[i]>0: sn11_ssrno[sn11_ssrb[i]] += 1

    sn1_ssrno  =  sn1_ssrno[np.unique(sn1_ssrb)]
    sn11_ssrno = sn11_ssrno[np.unique(sn1_ssrb)]
    sn11_ssrb = sn11_ssrb[sn11_ssrb>=0] 

    #---------------------------------------------------------------------
    #linear index (along diagonals of Michelogram) to rings
    # the number of Michelogram elements considered in projection calculations
    NLI2R_c = int(NRNG_c**2/2. + NRNG_c/2.)
    # if the whole scanner is used then account for the MRD and subtract 6 ring permutations
    if NRNG_c==NRNG: 
        NLI2R_c -= 6

    li2r   = np.zeros((NLI2R_c,2), dtype=np.int8)
    #the same as above but to sinos in span-11
    li2sn  = np.zeros((NLI2R_c,2), dtype=np.int16)
    li2sn1  = np.zeros((NLI2R_c,2), dtype=np.int16)
    li2rng = np.zeros((NLI2R_c,2), dtype=np.float32)
    #...to number of sinos (nos)
    li2nos = np.zeros((NLI2R_c), dtype=np.int8)

    dli = 0
    for ro in range(0, NRNG_c):
        # selects the sub-Michelogram of the whole Michelogram
        strt = NRNG*(ro+Cnt['RNG_STRT']) + Cnt['RNG_STRT']
        stop = (Cnt['RNG_STRT']+NRNG_c)*NRNG
        step = NRNG+1

        for li in range(strt, stop, step): #goes along a diagonal started in the first row at r2o
            #from the linear indexes of Michelogram get the subscript indexes
            r1 = li/NRNG
            r0 = li - r1*NRNG
            #avoid case when RD>MRD
            if (Msn[r1,r0])<0:
                continue
            # li2r[0, dli] = r0
            # li2r[1, dli] = r1
            # #--            
            # li2rng[0, dli] = rng[r0,0];
            # li2rng[1, dli] = rng[r1,0];
            # #-- 
            # li2sn[0, dli] = Msn[r0,r1]
            # li2sn[1, dli] = Msn[r1,r0]

            li2r[dli,0] = r0
            li2r[dli,1] = r1
            #--            
            li2rng[dli,0] = rng[r0,0]
            li2rng[dli,1] = rng[r1,0]
            #-- 
            li2sn[dli, 0] = Msn[r0,r1]
            li2sn[dli, 1] = Msn[r1,r0]

            li2sn1[dli, 0] = Msn1[r0,r1]
            li2sn1[dli, 1] = Msn1[r1,r0]

            # li2sn[0, dli] = Msn[r1,r0]
            # li2sn[1, dli] = Msn[r0,r1]
            #--
            li2nos[dli] = Mnos[r1,r0]
            #--
            dli += 1
    # print 'i> number of diagonal indeces (in Michelogram) accounted for:', dli
    #---------------------------------------------------------------------


    axLUT = {'li2rno':li2r, 'li2sn':li2sn, 'li2sn1':li2sn1, 'li2nos':li2nos, 'li2rng':li2rng, 
             'sn1_rno':sn1_rno, 'sn1_ssrb':sn1_ssrb, 'sn1_sn11':sn1_sn11, 'sn1_sn11no':sn1_sn11no,
             'sn11_ssrb':sn11_ssrb, 'sn1_ssrno':sn1_ssrno, 'sn11_ssrno':sn11_ssrno, 
             'Msn11':Msn, 'Msn1':Msn1, 'Mnos':Mnos, 'rng':rng}
    
    if Cnt['VERBOSE']:
        print 'i> axial LUTs done.'
    
    return axLUT


#=================================================================================================
def sino2ssr(sino, axLUT, Cnt):

    if Cnt['SPN']==1:
        slut = axLUT['sn1_ssrb']
        snno = Cnt['NSN1']
    elif Cnt['SPN']==11:
        slut = axLUT['sn11_ssrb']
        snno = Cnt['NSN11']
    else:
        print 'e> unrecognised span! (it is', Cnt['SPN'], ')'
        return None

    ssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)

    for i in range(snno):
        ssr[slut[i],:,:] += sino[i,:,:]

    return ssr

#=================================================================================================
def reduce_rings(pars, rs=0, re=64):
    ''' Reduce the axial rings for faster reconstructions, particularly simulations.
        This function customises axial FOV for reduced rings in range(rs,re).
        Note it only works in span-1 and ring re is not included in the reduced rings.
        Arguments:
        pars -- scanner parameters: constants, LUTs
        rs -- start ring
        re -- end ring (not included in the resulting reduced rings)
    '''

    Cnt = pars['Cnt']
    axLUT = pars['axLUT']

    pars['Cnt']['SPN'] = 1
    # select the number of sinograms for the number of rings
    # RNG_STRT is included in detection
    # RNG_END is not included in detection process
    pars['Cnt']['RNG_STRT'] = rs
    pars['Cnt']['RNG_END']  = re
    # now change the voxels dims too
    vz0 = 2*pars['Cnt']['RNG_STRT']
    vz1 = 2*(pars['Cnt']['RNG_END']-1)
    # number of axial voxels
    pars['Cnt']['rSO_IMZ'] = vz1-vz0+1
    pars['Cnt']['rSZ_IMZ'] = vz1-vz0+1
    # axial voxel size for scatter (mu-map and emission image)
    # pars['Cnt']['SS_IMZ'] = pars['Cnt']['rSG_IMZ']
    # number of rings customised for the given ring range (only optional in span-1)
    rNRNG = pars['Cnt']['RNG_END'] - pars['Cnt']['RNG_STRT']
    pars['Cnt']['rNRNG'] = rNRNG
    # number of reduced sinos in span-1
    rNSN1 = rNRNG**2
    pars['Cnt']['rNSN1'] = rNSN1
    # correct for the limited max. ring difference in the full axial extent. 
    # don't use ring range (1,63) as for this case no correction
    if rNRNG==64:  rNSN1 -= 12
    # apply the new ring subset to axial LUTs
    raxLUT = axial_lut(pars['Cnt'])
    # michelogram for reduced rings in span-1
    Msn1_c = raxLUT['Msn1']
    # michelogram for full ring case in span-1
    Msn1 = np.copy(pars['axLUT']['Msn1'])
    # from full span-1 sinogram index to reduced rings sinogram index
    rlut = np.zeros(rNSN1, dtype=np.int16)
    rlut[Msn1_c[Msn1_c>=0]] = Msn1[Msn1_c>=0]
    raxLUT['rLUT'] = rlut
    pars['axLUT'] = raxLUT


#=================================================================================================
def transaxial_lut(Cnt):
    "Creates a template 2D sino with gaps represented by 0; othterwise any valid bin is"
    "represented as 1."
    "Also creates linear index for the whole sino with only valid bins.  Angle index of"
    "of the sino is used as the primary index (fast changing)."

    # #---visualisation of the crystal ring in trasaxial view
    # p = 8 #pixel density of the visualiastion
    # VISXY = 344*p
    # T = np.zeros((VISXY,VISXY), dtype=np.float32)
    # #---

    #--- crystal coordinates transaxially
    #block width
    bw = 3.209
    #block gap [cm]
    dg = 0.474
    NTBLK = 56
    alpha = 0.1122  #2*pi/NTBLK
    crs = np.zeros((4,Cnt['NCRS']), dtype=np.float32)
    # phi angle points in the middle and is used for obtaining the normal of detector block
    phi = 0.5*pi - alpha/2 -0.001
    for bi in range(NTBLK):
        #-tangent point (ring against detector block)
        # ye = RE*np.sin(phi)
        # xe = RE*np.cos(phi)
        y  =  Cnt['RE']*np.sin(phi)
        x  =  Cnt['RE']*np.cos(phi)
        #-vector for the face of crystals
        pv  = np.array([-y, x])
        pv /= np.sum(pv**2)**.5
        #update phi for next block
        phi -= alpha
        #-end block points
        xcp = x + (bw/2)*pv[0]
        ycp = y + (bw/2)*pv[1]
        # u = int( .5*VISXY + np.floor(xcp/(Cnt['SO_VXY']/p)) )
        # v = int( .5*VISXY - np.ceil (ycp/(Cnt['SO_VXY']/p)) )
        # T[v,u] = 5
        for n in range(1,9):
            c = bi*9 +n-1
            crs[0,c] = xcp
            crs[1,c] = ycp
            xc = x + (bw/2-n*bw/8)*pv[0]
            yc = y + (bw/2-n*bw/8)*pv[1]
            crs[2,c] = xc
            crs[3,c] = yc
            xcp = xc
            ycp = yc

    # cij    - a square matrix of crystals in coincidence (transaxially)
    # crsri  - indexes of crystals with the gap crystals taken out (therefore reduced)
    # aw2sn  - LUT array [AW x 2] translating linear index into a 2D sinogram with dead LOR (gaps)
    # aw2ali - LUT from linear index of 2D full sinogram with gaps and bin-driven to
    #          linear index without gaps and angle driven
    # msino  - 2D sinogram with gaps marked (0). like a mask.
    Naw, s2cAll, crsri, cij, aw2sn, aw2ali, msino = mmr_auxe.txlut( Cnt )
    s2cF = s2cAll[0]
    s2c  = s2cAll[1]
    s2cr = s2cAll[2]
    c2sF = s2cAll[3]
    cr2s = s2cAll[4]

    txLUT = {'cij':cij, 'crs':crs, 'crsri':crsri, 'msino':msino, 'aw2sn':aw2sn,
             'aw2ali':aw2ali, 's2c':s2c, 's2cr':s2cr, 's2cF':s2cF, 'Naw':Naw,
             'c2sF':c2sF, 'cr2s':cr2s}
    return txLUT



#=================================================================================================
# Explore files in folder with raw PET/MR data
#-------------------------------------------------------------------------------------------------
def get_npfiles(dfile, datain, v):
    if v:
        print '--------------------------------------------------------------------------'
        print 'i> file:', dfile

    # pCT mu-map
    if os.path.basename(dfile)=='mumap_pCT.npy':
        datain['mumapCT'] = dfile
        if v: print 'mu-map for the object.'

    # DICOM UTE/Dixon mu-map
    if os.path.basename(dfile)=='mumap-from-DICOM.npy':
        datain['mumapNPY'] = dfile
        if v: print 'mu-map for the object.'

    if os.path.basename(dfile)=='hmumap.npy':
        datain['hmumap'] = dfile
        if v: print 'mu-map for hardware.'

    if os.path.basename(dfile)[:8]=='sinos_s1':
        datain['sinos'] = dfile
        if v: print 'prompt sinogram data.'

    # if os.path.basename(dfile)[:9]=='sinos_s11':
    #     datain['sinos11'] = dfile
    #     if v: print 'prompt sinogram data in span-11.'


def get_niifiles(dfile, datain, v):
    if v:
        print '--------------------------------------------------------------------------'
        print 'i> file:', dfile


    #> NIfTI file of converted MR-based mu-map from DICOMs
    if os.path.basename(dfile).split('.nii')[0]=='mumap-from-DICOM':
        datain['mumapNII'] = dfile
        if v: print 'mu-map for the object.'


    #> NIfTI file of pseudo CT
    fpct = glob.glob( os.path.join(os.path.dirname(dfile), '*_synth.nii*') )
    if len(fpct)>0:
        datain['pCT'] = fpct[0]
        if v: print 'i> pseudoCT of the object.'

    fpct = glob.glob( os.path.join(os.path.dirname(dfile), '*_p[cC][tT].nii*') )
    if len(fpct)>0:
        datain['pCT'] = fpct[0]
        if v: print 'i> pseudoCT of the object.'

    #MR T1 
    fmri = glob.glob( os.path.join(os.path.dirname(dfile), '[tT]1*.nii*') )
    if len(fmri)==1:
        bnm = os.path.basename(fmri[0]).lower()
        if not 'giflabels' in bnm and not 'parcellation' in bnm \
        and not 'pct' in bnm and not 'n4bias' in bnm:
            datain['T1nii'] = fmri[0]
            if v: print 'i> NIfTI for T1w of the object.'
    elif len(fmri)>1:
        for fg in fmri:
            bnm = os.path.basename(fg).lower()
            if not 'giflabels' in bnm and not 'parcellation' in bnm\
            and not 'pct' in bnm and not 'n4bias' in bnm:
                if 'preferred' in bnm:
                    datain['T1nii'] = fg
                elif 'usable' in bnm:
                    datain['T1nii_2'] = fg

    #MR T1 N4bias-corrected
    fmri = glob.glob( os.path.join(os.path.dirname(dfile), '[tT]1*[nN]4bias*.nii*') )
    if len(fmri)==1:
        bnm = os.path.basename(fmri[0]).lower()
        if not 'giflabels' in bnm and not 'parcellation' in bnm \
        and not 'pct' in bnm:
            datain['T1N4'] = fmri[0]
            if v: print 'i> NIfTI for T1w of the object.'
    elif len(fmri)>1:
        for fg in fmri:
            bnm = os.path.basename(fg).lower()
            if not 'giflabels' in bnm and not 'parcellation' in bnm\
            and not 'pct' in bnm:
                if 'preferred' in bnm:
                    datain['T1N4'] = fg
                elif 'usable' in bnm:
                    datain['T1N4_2'] = fg


    #T1w corrected
    fbc = glob.glob( os.path.join(os.path.dirname(dfile), '*gifbc.nii*') )
    if len(fbc)==1:
        datain['T1bc'] = fbc[0]
        if v: print 'i> NIfTI for bias corrected T1w of the object:', fbc[0]
    fbc = glob.glob( os.path.join(os.path.dirname(dfile), '*[tT]1*BiasCorrected.nii*') )
    if len(fbc)==1:
        datain['T1bc'] = fbc[0]
        if v: print 'i> NIfTI for bias corrected T1w of the object:', fbc[0]

    #T1-based labels after parcellation
    flbl = glob.glob( os.path.join(os.path.dirname(dfile), '*giflabels.nii*') )
    if len(flbl)==1:
        datain['T1lbl'] = flbl[0]
        if v: print 'i> NIfTI for regional parcellations of the object:', flbl[0]
    flbl = glob.glob( os.path.join(os.path.dirname(dfile), '*[tT]1*[Pp]arcellation.nii*') )
    if len(flbl)==1:
        datain['T1lbl'] = flbl[0]
        if v: print 'i> NIfTI for regional parcellations of the object:', flbl[0]

    #reconstructed emission data without corrections, minimum 2 osem iter
    fpct = glob.glob( os.path.join(os.path.dirname(dfile), '*__ACbed.nii*') )
    if len(fpct)>0:
        datain['em_nocrr'] = fpct[0]
        if v: print 'i> pseudoCT of the object.'

    #reconstructed emission data with corrections, minimum 3 osem iter
    fpct = glob.glob( os.path.join(os.path.dirname(dfile), '*QNT*.nii*') )
    if len(fpct)>0:
        datain['em_crr'] = fpct[0]
        if v: print 'i> pseudoCT of the object.'


# =======================================================================================
def get_dicoms(dfile, datain, Cnt):

    v = Cnt['VERBOSE']
    if v: 
        print '--------------------------------------------------------------------------'
        print 'i> file:', dfile

    d = dcm.dcmread(dfile)
    dcmtype = nimpa.dcminfo(d, verbose=Cnt['VERBOSE'])

    #> check if it is norm file
    if 'mmr' in dcmtype and 'norm' in dcmtype:
        if os.path.splitext(dfile)[-1].lower()=='.dcm':
            datain['nrm_dcm'] = dfile
            
            #> check if the binary file exists
            if os.path.isfile(dfile[:-4]+'.bf'):
                datain['nrm_bf'] = dfile[:-4]+'.bf'
            else:
                print 'e> file does not exists:', dfile[:-4]+'.bf'
        elif os.path.splitext(dfile)[-1].lower()=='.ima':
            datain['nrm_ima'] = dfile
            # extract the binary norm data from the IMA DICOM
            if [0x7fe1,0x1010] in d:
                nrm = d[0x7fe1,0x1010].value
            else:
                print 'e> could not find binary normalisation data in the IMA DICOM file.'
            # binary file name
            bf = os.path.splitext(dfile)[0]+'.bf'
            with open(bf, 'wb') as f:
                f.write(nrm)
            datain['nrm_bf'] = bf
            if Cnt['VERBOSE']: print 'i> saved component norm data to binary file:', bf

    #--- check if it is list-mode file     
    elif 'mmr' in dcmtype and 'list' in dcmtype:
        if os.path.splitext(dfile)[-1]=='.dcm':
            datain['lm_dcm'] = dfile
            #check if the binary file exists
            if os.path.isfile(dfile[:-4]+'.bf'):
                datain['lm_bf'] = dfile[:-4]+'.bf'
            else:
                print 'e> file does not exists:', dfile[:-4]+'.bf'
        elif os.path.splitext(dfile)[-1].lower()=='.ima':
            datain['lm_ima'] = dfile
            # extract the binary list-mode data from the IMA DICOM if it does not exist already
            # binary file name
            bf = os.path.splitext(dfile)[0] + '.bf'
            if [0x7fe1,0x1010] in d and not os.path.isfile(bf):
                lm = d[0x7fe1,0x1010].value
                with open(bf, 'wb') as f:
                    f.write(lm)
                datain['lm_bf'] = bf
                if Cnt['VERBOSE']: print 'i> saved list-mode data to binary file:', bf
            elif os.path.isfile(bf):
                if Cnt['VERBOSE']: print 'i> the binary list-mode data was already extracted from the IMA DICOM file.'
                datain['lm_bf'] = bf
            else:
                print 'e> could not find binary list-mode data in the IMA DICOM file.'
                return None
        
        #> get info about the PET tracer being used
        lmhdr, csahdr = hdr_lm(datain, Cnt)

        #> if there is interfile header get the info from there
        if lmhdr!=None:
            f0 = lmhdr.find('isotope name')
        else:
            f0 = -1
        
        if f0>=0:
            f1 = f0+lmhdr[f0:].find('\n')
            #regular expression for the isotope symbol
            p = re.compile(r'(?<=:=)\s*\S*')
            # the name of isotope:
            istp = p.findall(lmhdr[f0:f1])[0]
            istp = istp.replace('-', '')
            Cnt['ISOTOPE'] = istp.strip()

        #> if no info in interfile header than look in the CSA header
        else:
            f0 = csahdr.find('RadionuclideCodeSequence')
            if f0<0: 
                print 'w> could not find isotope name.  enter manually into Cnt[''ISOTOPE'']'
                return None
            istp_coded = re.search('(?<=CodeValue:)\S*', csahdr[f0:f0+100]).group()
            if   istp_coded=='C-111A1':   Cnt['ISOTOPE'] = 'F18'
            elif istp_coded=='C-105A1':   Cnt['ISOTOPE'] = 'C11'
            elif istp_coded=='C-B1038':   Cnt['ISOTOPE'] = 'O15'
            elif istp_coded=='C-128A2':   Cnt['ISOTOPE'] = 'Ge68'
            elif istp_coded=='C-131A3':   Cnt['ISOTOPE'] = 'Ga68'
            else:   
                print 'w> could not find isotope name.  enter manually into Cnt[''ISOTOPE'']'
                return None
        #---

    # check if MR-based mu-map
    elif 'mumap' in dcmtype:
        datain['mumapDCM'] = os.path.dirname(dfile)
        if '#mumapDCM' not in datain:
            datain['#mumapDCM'] = 1
        else:
            datain['#mumapDCM'] += 1

    # check for MR T1w and T2w images
    elif 'mr' in dcmtype and 't1' in dcmtype:
        datain['T1DCM'] = os.path.dirname(dfile)
        if '#T1DCM' not in datain:
            datain['#T1DCM'] = 1
        else:
            datain['#T1DCM'] += 1

    elif 'mr' in dcmtype and 't2' in dcmtype:
        datain['T2DCM'] = os.path.dirname(dfile)
        if '#T2DCM' not in datain:
            datain['#T2DCM'] = 1
        else:
            datain['#T2DCM'] += 1

    # UTE's two sequences:
    elif 'mr' in dcmtype and 'ute2' in dcmtype:
        datain['UTE2'] = os.path.dirname(dfile)
        if '#UTE2' not in datain:
            datain['#UTE2'] = 1
        else:
            datain['#UTE2'] += 1

    elif 'mr' in dcmtype and 'ute1' in dcmtype:
        datain['UTE1'] = os.path.dirname(dfile)
        if '#UTE1' not in datain:
            datain['#UTE1'] = 1
        else:
            datain['#UTE1'] += 1

    
    if Cnt['VERBOSE']: print ''

#-------------------------------------------------------------------------------------------------
def explore_input(fldr, params, print_paths=False):

    # two ways of passing Cnt are here decoded
    if 'Cnt' in params:
        Cnt = params['Cnt']
    else:
        Cnt = params

    if not os.path.isdir(fldr):
        print 'e> provide a valid folder path for the data.'
        return

    #check for the availble data: list mode data, component-based norm and mu-maps
    # [dcm + bf] is one format of DICOM raw data; [ima] is another one used.
    # mu-map can be given from the scanner as an e.g., UTE-based, or pseudoCT through synthesis.
    datain = {'corepath' :''}

    datain['corepath'] = fldr

    for f in os.listdir(fldr):
        if os.path.isfile( pjoin(fldr,f) ):
            if f.lower().endswith(".dcm") or f.lower().endswith(".ima"):
                get_dicoms( pjoin(fldr,f), datain, Cnt)
            # elif f.lower().endswith(".bf"):
            #     get_bf( pjoin(fldr,f), datain, Cnt)
            elif f.endswith(".npy") or f.endswith(".dic"):
                get_npfiles( pjoin(fldr,f), datain, Cnt['VERBOSE'])
            elif f.endswith(".nii.gz") or f.endswith(".nii"):
                get_niifiles( pjoin(fldr,f), datain, Cnt['VERBOSE'])

        elif os.path.isdir(pjoin(fldr,f)):
            #go one level into subfolder
            sfldr = pjoin(fldr,f)
            for sf in os.listdir(sfldr):
                if os.path.isfile( pjoin(sfldr, sf) ):
                    if sf.lower().endswith(".dcm") or sf.lower().endswith(".ima"):
                        get_dicoms(   pjoin(sfldr,sf), datain, Cnt )
                    elif sf.endswith(".nii.gz") or sf.endswith(".nii"):
                        get_niifiles( pjoin(sfldr,sf), datain, Cnt['VERBOSE'])
                    elif sf.endswith(".npy") or sf.endswith(".dic"):
                        get_npfiles(  pjoin(sfldr,sf), datain, Cnt['VERBOSE'])
    
    if print_paths:           
        print '--------------------------------------------------'
        for x in datain:
            print x,':',datain[x]
        print '--------------------------------------------------'

    return datain

#=====================================================================================
def putgaps(s, txLUT, Cnt):

    #number of sino planes (2D sinos) depends on the span used
    if Cnt['SPN']==1:
        # number of rings calculated for the given ring range (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent (don't use ring range (1,63) as for this case no correction) 
        if NRNG_c==64:
            nsinos -= 12

    elif Cnt['SPN']==11:
        nsinos = Cnt['NSN11']

    #preallocate sino with gaps
    sino = np.zeros((Cnt['NSANGLES'], Cnt['NSBINS'], nsinos), dtype=np.float32)
    #fill the sino with gaps
    mmr_auxe.pgaps(sino, s.astype(np.float32), txLUT, Cnt)
    sino = np.transpose(sino, (2,0,1))

    return sino.astype(s.dtype)

def remgaps(sino, txLUT, Cnt):

    # number of sino planes (2D sinos) depends on the span used
    nsinos = sino.shape[0]

    #preallocate output sino without gaps, always in float
    s = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
    #fill the sino with gaps
    mmr_auxe.rgaps(s, sino.astype(np.float32), txLUT, Cnt)

    # return in the same data type as the input sino
    return s.astype(sino.dtype)
#================================================================================================


def mmrinit():
    # get the constants for the mMR
    Cnt = resources.get_mmr_constants()

    # transaxial look up tables
    txLUT = transaxial_lut(Cnt)
    Cnt['Naw'] = txLUT['Naw']

    # axial look up tables
    axLUT = axial_lut(Cnt)

    return Cnt, txLUT, axLUT

def mMR_params():
    '''
    get all scanner parameters in one dictionary
    '''
    Cnt, txLUT, axLUT = mmrinit()
    return {'Cnt':Cnt, 'txLUT':txLUT, 'axLUT':axLUT}


# init the package folder
import mmrhist
from mmrhist import frame_position
from mmrhist import auxilary_frames
from mmrhist import dynamic_timings
from mmrhist import split_frames
from mmrhist import draw_frames
from mmrhist import get_time_offset

from pviews import video_frm, video_dyn


"""Simulations for image reconstruction with recommended reduced axial field of view"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018, University College London"

import numpy as np

from niftypet import nimpa
# from niftypet.nipet.prj import mmrprj

import mmrprj
import mmrrec
import petprj

from niftypet.nipet import mmraux
from niftypet.nipet.img import mmrimg

from tqdm.auto import trange


def simulate_sino(
        petim,
        ctim,
        scanner_params,
        simulate_3d = False,
        slice_idx=-1,
        mu_input = False):
    ''' Simulate the measured sinogram with photon attenuation.
        Arguments:
        petim -- the input PET image based on which the emission sinogram is found
        ctim -- CT image, in register with PET and the same dimensions, is used 
            for estimating the attenuation factors, which are then applied to 
            simulate emission sinogram with realistic photon attenuation. 
        slice_idx -- chosen 2D slice out of the 3D image for the fast simulation.
        scanner_params -- scanner parameters containing scanner constants and
            axial and transaxial look up tables (LUTs)
        mu_input -- if True, the values are representative of a mu-map in [1/cm],
            otherwise it represents the CT in [HU].
    '''

    #> decompose the scanner constants and LUTs for easier access
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if petim.shape != ctim.shape:
        raise ValueError('The shapes of the PET and CT images are inconsistent.')


    if simulate_3d:

        if petim.ndim!=3 \
                or petim.shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            raise ValueError(
                'The input image shape does not match the scanner image size.')

        if petim.max()>200:
            print 'w> the PET image may have too large intensities for robust simulation.'

    else:

        #> 2D case with reduced rings
        if len(petim.shape)==3:
            
            # make sure that the shape of the input image matches the image size of the scanner
            if petim.shape[1:]!=(Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')

            # pick the right slice index (slice_idx) if not given or mistaken
            if slice_idx<0:
                print 'w> the axial index <slice_idx> is chosen to be in the middle of axial FOV.'
                slice_idx = petim.shape[0]/2
            if slice_idx>=petim.shape[0]:
                raise ValueError('The axial index for 2D slice selection is outside the image.')

        elif len(petim.shape)==2:
            
            # make sure that the shape of the input image matches the image size of the scanner
            if petim.shape != (Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')

            petim.shape = (1,) + petim.shape
            ctim.shape  = (1,) + ctim.shape
            slice_idx = 0

        if not 'rSZ_IMZ' in Cnt:
            raise ValueError('Missing reduced axial FOV parameters.')

    


    # import pdb; pdb.set_trace()

    #--------------------
    if mu_input:
        mui = ctim
    else:
        #> get the mu-map [1/cm] from CT [HU] 
        mui = nimpa.ct2mu(ctim)
    
    #> get rid of negative values
    mui[mui<0] = 0
    #--------------------

    if simulate_3d:
        
        rmu = mui
        rpet = petim

    else:

        #> 2D case with reduced rings
        #--------------------
        #> create a number of slices of the same chosen image slice for reduced (fast) 3D simulation
        rmu = mui[slice_idx,:,:]
        rmu.shape = (1,) + rmu.shape
        rmu = np.repeat(rmu, Cnt['rSZ_IMZ'], axis=0)
        #--------------------

        #--------------------
        #> form a short 3D image of the same emission image slice
        rpet = petim[slice_idx,:,:].copy()
        rpet.shape = (1,) + rpet.shape
        rpet = np.repeat(rpet, Cnt['rSZ_IMZ'], axis=0)
        #--------------------

    #> forward project the mu-map to obtain attenuation factors
    attsino = mmrprj.frwd_prj(rmu,  scanner_params, attenuation=True)

    #> forward project the PET image to obtain non-attenuated emission sino
    emisino = mmrprj.frwd_prj(rpet, scanner_params, attenuation=False)

    #> return the simulated emission sino with photon attenuation 
    return attsino*emisino



def simulate_recon(
        measured_sino,
        ctim,
        scanner_params,
        simulate_3d = False,
        nitr = 60,
        slice_idx = -1,
        randoms=None,
        mu_input = False,
        msk_radius = 29.
    ):

    ''' Reconstruct PET image from simulated input data using the EM-ML algorithm.
        Arguments:
        measured_sino -- simulated emission data with photon attenuation
        ctim -- either a 2D CT image or a 3D CT image from which a 2D slice is chosen (slice_idx) for estimation
            of the attenuation factors
        slice_idx -- index to extract one 2D slice for this simulation if input image is 3D
        nitr -- number of iterations used for the EM-ML reconstruction algorithm
        scanner_params -- scanner parameters containing scanner constants and
            axial and transaxial look up tables (LUTs)
        randoms[=None] -- possibility of using randoms and scatter events in the simulation  
    '''

    #> decompose the scanner constants and LUTs for easier access
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    
    if simulate_3d:

        if ctim.ndim!=3 \
                or ctim.shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            raise ValueError(
                'The CT/mu-map image does not match the scanner image shape.')
    else:

        #> 2D case with reduced rings
        if len(ctim.shape)==3:
            
            # make sure that the shape of the input image matches the image size of the scanner
            if ctim.shape[1:]!=(Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')

            # pick the right slice index (slice_idx) if not given or mistaken
            if slice_idx<0:
                print 'w> the axial index <slice_idx> is chosen to be in the middle of axial FOV.'
                slice_idx = ctim.shape[0]/2
            if slice_idx>=ctim.shape[0]:
                raise ValueError('The axial index for 2D slice selection is outside the image.')

        elif len(ctim.shape)==2:
            
            # make sure that the shape of the input image matches the image size of the scanner
            if ctim.shape != (Cnt['SO_IMY'], Cnt['SO_IMX']):
                raise ValueError('The input image shape for x and y does not match the scanner image size.')

            ctim.shape  = (1,) + ctim.shape
            slice_idx = 0

        if not 'rSZ_IMZ' in Cnt:
            raise ValueError('Missing reduced axial FOV parameters.')


    #--------------------
    if mu_input:
        mui = ctim
    else:
        #> get the mu-map [1/cm] from CT [HU] 
        mui = nimpa.ct2mu(ctim)
    
    #> get rid of negative values
    mui[mui<0] = 0
    #--------------------


    if simulate_3d:

        rmu = mui

        #> number of axial sinograms
        nsinos = Cnt['NSN11']

    else:
        #--------------------
        #> create a number of slides of the same chosen image slice for reduced (fast) 3D simulation
        rmu = mui[slice_idx,:,:]
        rmu.shape = (1,) + rmu.shape
        rmu = np.repeat(rmu, Cnt['rSZ_IMZ'], axis=0)
        #--------------------

        #> number of axial sinograms
        nsinos = Cnt['rNSN1']

    # import pdb; pdb.set_trace()


    #> attenuation factor sinogram
    attsino = mmrprj.frwd_prj(rmu,  scanner_params, attenuation=True, dev_out=True)

    nrmsino = np.ones(attsino.shape, dtype=np.float32)

    #> randoms and scatter put together
    if randoms==None:
        rndsct = 1e-5*np.ones((Cnt['Naw'], nsinos), dtype=np.float32)
    else:
        rndsct = randoms


    

    if simulate_3d:

        if Cnt['VERBOSE']:
            print '\n>------ OSEM (', nitr,  ') -------\n'

        # measured sinogram in GPU-enabled shape
        psng = mmraux.remgaps(measured_sino.astype(np.uint16), txLUT, Cnt)

        #> mask for reconstructed image.  anything outside it is set to zero
        msk = mmrimg.get_cylinder(Cnt, rad=msk_radius, xo=0, yo=0, unival=1, gpu_dim=True)>0.9

        #> init image
        eimg = np.ones((Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)

        #------------------------------------
        Sn = 14 # number of subsets
        #-get one subset to get number of projection bins in a subset
        Sprj, s = mmrrec.get_subsets14(0,scanner_params)
        Nprj = len(Sprj)
        
        #> init subset array and sensitivity image for a given subset
        sinoTIdx = np.zeros((Sn, Nprj+1), dtype=np.int32)
        
        #> init sensitivity images for each subset
        sim = np.zeros((Sn, Cnt['SZ_IMY'], Cnt['SZ_IMX'], Cnt['SZ_IMZ']), dtype=np.float32)
        
        for n in range(Sn):
            sinoTIdx[n,0] = Nprj #first number of projection for the given subset
            sinoTIdx[n,1:], s = mmrrec.get_subsets14(n,scanner_params)
            #> sensitivity image
            petprj.bprj(
                sim[n,:,:,:],
                attsino[sinoTIdx[n,1:],:],
                txLUT,
                axLUT,
                sinoTIdx[n,1:],
                Cnt)
        #-------------------------------------

        for k in trange(nitr, disable=not Cnt['VERBOSE'], desc="OSEM"):
            petprj.osem(
                eimg,
                msk,
                psng,
                rndsct,
                rndsct,
                nrmsino,
                attsino,
                sim,
                txLUT,
                axLUT,
                sinoTIdx,
                Cnt)

        eim = mmrimg.convert2e7(eimg, Cnt)



    else:

        #> estimated image, initialised to ones
        eim = np.ones(rmu.shape, dtype=np.float32)

        msk = mmrimg.get_cylinder(Cnt, rad=msk_radius, xo=0, yo=0, unival=1, gpu_dim=False)>0.9

        #> sensitivity image for the EM-ML reconstruction
        sim = mmrprj.back_prj(attsino, scanner_params)

        for i in range(nitr):
            if Cnt['VERBOSE']: print '>---- EM iteration:', i
            #> remove gaps from the measured sinogram
            #> then forward project the estimated image
            #> after which divide the measured sinogram by the estimated sinogram (forward projected)
            crrsino = mmraux.remgaps(measured_sino, txLUT, Cnt) / \
                        (mmrprj.frwd_prj(eim, scanner_params, dev_out=True) + rndsct)

            #> back project the correction factors sinogram
            bim = mmrprj.back_prj(crrsino, scanner_params) 

            #> divide the back-projected image by the sensitivity image
            bim[msk] /= sim[msk]
            bim[~msk] = 0

            #> update the estimated image and remove NaNs 
            eim *= msk*bim
            eim[np.isnan(eim)] = 0



    return eim


import numpy as np
import math
import os
import sys
import scipy.ndimage as ndi
import csv
import zipfile
import pydicom as dcm

import re
import nibabel as nib
from subprocess import call
import glob
import platform

#--------------
import pycurl
import json
from StringIO import StringIO
from urllib import urlencode
from datetime import datetime
#--------------

from niftypet import nimpa

# DICOM extensions
dcm_ext = ('dcm', 'DCM', 'ima', 'IMA')


# # version of matplotlib to int
# import matplotlib.pyplot as plt
# import matplotlib
# mplv =  matplotlib.__version__.split('.')
# mplv = int(mplv[0])*10 + int(mplv[1])

# -----------------------------
def create_dir(pth):
    if not os.path.exists(pth):    
        os.makedirs(pth)
# -----------------------------

# -----------------------------
def time_stamp():
    now    = datetime.now()
    nowstr = str(now.year)+'-'+str(now.month)+'-'+str(now.day)+' '+str(now.hour)+':'+str(now.minute)
    return nowstr
# -----------------------------


#----------------------------------------------------------------------------------------------------------
def get_xnatList(xnaturi, cookie='', usrpwd=''):
    buffer = StringIO()
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    elif usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.WRITEDATA, buffer)
    c.perform()
    c.close()
    # convert to json dictionary in python
    outjson = json.loads( buffer.getvalue() )
    return outjson['ResultSet']['Result']

def get_xnat(xnaturi, frmt='json', cookie='', usrpwd=''):
    buffer = StringIO()
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    if usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.WRITEDATA, buffer)
    c.perform()
    c.close()
    # convert to json dictionary in python
    if frmt=='':
        output = buffer.getvalue()
    elif frmt=='json':
        output = json.loads( buffer.getvalue() )
    return output

def get_xnatFile(xnaturi, fname, cookie='', usrpwd=''):
    try:
        fn = open(fname, 'wb')
        c = pycurl.Curl()
        if cookie:
            c.setopt(pycurl.COOKIE, cookie)
        else:
            c.setopt(c.USERPWD, usrpwd)
        c.setopt(pycurl.SSL_VERIFYPEER, 0)
        c.setopt(pycurl.SSL_VERIFYHOST, 0)
        c.setopt(c.VERBOSE, 0)
        c.setopt(c.URL, xnaturi )
        c.setopt(c.WRITEDATA, fn)
        c.setopt(pycurl.FOLLOWLOCATION, 0)
        c.setopt(pycurl.NOPROGRESS, 0)
        c.perform()
        c.close()
        fn.close()
    except pycurl.error as pe:
        print ' '
        print '=============================================================='
        print 'e> pycurl error:', pe
        print '=============================================================='
        print ' '
        print 'w> no data.'
        return -1
    else:
        print '--------'
        print 'i> pycurl download done.'
        print '--------'
    return 0
#----------------------------------------------------------------------------------------------------------


#----------------------------------------------------------------------------------------------------------
def put_xnat(xnaturi, cookie='', usrpwd=''):
    """e.g., create a container"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    elif usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.CUSTOMREQUEST, 'PUT')
    c.perform()
    c.close()

def del_xnat(xnaturi, cookie='', usrpwd=''):
    """e.g., create a container"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    elif usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.CUSTOMREQUEST, 'DELETE')
    c.perform()
    c.close()

def post_xnat(xnaturi, post_data, verbose=0, PUT=False,  cookie='', usrpwd=''):
    buffer = StringIO()
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    elif usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(c.USERPWD, usrpwd)
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(c.VERBOSE, verbose)
    c.setopt(c.URL, xnaturi )
    if PUT: c.setopt(c.CUSTOMREQUEST, 'PUT')
    c.setopt(c.POSTFIELDS, post_data)
    c.setopt(c.WRITEFUNCTION, buffer.write)
    c.perform()
    c.close()
    return buffer.getvalue()

def put_xnatFile(xnaturi, filepath, cookie='', usrpwd=''):
    """upload file to xnat server"""
    c = pycurl.Curl()
    if cookie:
        c.setopt(pycurl.COOKIE, cookie)
    elif usrpwd:
        c.setopt(c.USERPWD, usrpwd)
    else:
        raise NameError('Session ID or username:password are not given')
    c.setopt(pycurl.SSL_VERIFYPEER, 0)
    c.setopt(pycurl.SSL_VERIFYHOST, 0)
    c.setopt(pycurl.NOPROGRESS, 0)
    c.setopt(c.VERBOSE, 0)
    c.setopt(c.URL, xnaturi )
    c.setopt(c.HTTPPOST, [('fileupload', (c.FORM_FILE, filepath,)),])
    c.perform()
    c.close()
#----------------------------------------------------------------------------------------------------------



#----------------------------------------------------------------------------------------------------------
def put_xnatPetMrRes(usrpwd, xnatsbj, sbjix, lbl, frmt, fpth):
    # get the experiment id
    expt = get_xnatList(usrpwd, xnatsbj+'/' +sbjix+ '/experiments?xsiType=xnat:petmrSessionData&format=json&columns=ID')
    # prepare the uri
    xnaturi = xnatsbj+'/' +sbjix+ '/experiments/' + expt[0]['ID'] + '/resources/' +lbl+ '?xsi:type=xnat:resourceCatalog&format='+frmt
    xnat_put(usrpwd, xnaturi)
    # upload
    xnaturi = xnatsbj+'/' +sbjix+ '/experiments/' + expt[0]['ID'] + '/resources/' +lbl+ '/files'
    xnat_upload(usrpwd, xnaturi, fpth)
#----------------------------------------------------------------------------------------------------------




def getscan(
        sbjix,
        expt,
        xc,
        scan_types = '',
        scan_ids = '',
        cookie = '',
        outpath = '',
        fcomment = '',
        #close_session=True,
        ):


    if not cookie:
        sessionID = nipet.xnat.post_xnat(xc['url']+'/data/JSESSIONID', '', usrpwd=xc['usrpwd'])
        cookie = 'JSESSIONID='+sessionID

    #> output dictionary
    out = {}
    out['cookie'] = cookie

    if outpath=='':
        if os.path.isdir(xc['opth']):
            opth = xc['opth']
        else:
            if platform.system() in ['Linux', 'Darwin']:
                opth = os.path.join(os.path.expanduser('~'), 'xnat_scans')
            elif platform.system() == 'Windows' :
                opth = os.path.join(os.getenv('LOCALAPPDATA'), 'xnat_scans')
            else:
                raise IOError('e> unknown system and no output folder provided!')
    else:
        opth = outpath


    scans = get_xnatList(
        xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans',
        cookie=cookie
    )

    all_scan_types = [(s['type'],s['quality'],s['ID']) for s in scans]


    picked_scans = []
    if scan_types:
        for st in scan_types:
            picked_scans.extend([s for s in all_scan_types if st in s[0]])
    
    elif scan_ids:
        for si in scan_ids:
            picked_scans.extend([s for s in all_scan_types if si == s[2]])
    else:
        raise ValueError('e> unspecified scans to download!')


    for scn in picked_scans:

        stype   = str(scn[0])
        quality = str(scn[1])
        sid     = str(scn[2])

        s_type_id = sid+'_'+stype

        entries = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans/'+sid+'/resources',
            cookie=cookie
        )

        for e in entries:

            if e['format'] in ['DICOM', 'NIFTI']:

                out[s_type_id] = []

                files = get_xnatList(
                        xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] \
                            + '/scans/'+sid+'/resources/'+ e['format']+ '/files',
                        cookie=cookie
                        )

                #> scan path
                spth = os.path.join( opth,  s_type_id+'__'+quality)
                nimpa.create_dir(spth)

                #> download all files
                for i in range(len(files)):
                    
                    fname = 'scan-'+s_type_id+'__'+quality+fcomment\
                            +'.'+files[i]['Name'].split('.',1)[-1]

                    status = get_xnatFile(
                        xc['url']+files[i]['URI'],
                        os.path.join(spth, fname),
                        cookie=cookie)

                    if status<0:
                        print 'e> no scan data for', scntype
                    else:
                        out[s_type_id].append(os.path.join(spth, fname))
                
                if len(files)<1: 
                    print 'e> no scan data for', scntype


def getresources(
        rfiles,
        xc,
        outpath = '',
        cookie = '',
        ):


    if not cookie:
        sessionID = nipet.xnat.post_xnat(xc['url']+'/data/JSESSIONID', '', usrpwd=xc['usrpwd'])
        cookie = 'JSESSIONID='+sessionID

    #> output dictionary
    out = {}
    out['cookie'] = cookie

    if outpath=='':
        if os.path.isdir(xc['opth']):
            opth = xc['opth']
        else:
            if platform.system() in ['Linux', 'Darwin']:
                opth = os.path.join(os.path.expanduser('~'), 'xnat_scans')
            elif platform.system() == 'Windows' :
                opth = os.path.join(os.getenv('LOCALAPPDATA'), 'xnat_scans')
            else:
                raise IOError('e> unknown system and no output folder provided!')
    else:
        opth = outpath


    for i in range(len(rfiles)):

        #> check if the file is already downloaded:
        if  os.path.isfile ( os.path.join(opth, rfiles[i]['Name']) ) and \
            str(os.path.getsize(os.path.join(opth, rfiles[i]['Name'])))==rfiles[i]['Size']:
            
            print 'i> file of the same size,',rfiles[i]['Name'], 'already exists: skipping download.'
            
            if '.dcm' in rfiles[i]['Name'].lower():
                if 'dcm' not in out: out['dcm'] = []
                out['dcm'].append(os.path.join(opth, rfiles[i]['Name']))
            elif '.bf' in rfiles[i]['Name'].lower():
                if 'bf' not in out: out['bf'] = []
                out['bf'].append(os.path.join(opth, rfiles[i]['Name']))
            elif '.ima' in rfiles[i]['Name'].lower():
                if 'ima' not in out: out['ima'] = []
                out['ima'].append(os.path.join(opth, rfiles[i]['Name']))
        
        else:
            status = get_xnatFile(
                xc['url']+rfiles[i]['URI'], os.path.join(opth, rfiles[i]['Name']),
                cookie = cookie
            )
            if status<0:
                print 'e> error downloading:', fcomment
            else:
                if '.dcm' in rfiles[i]['Name'].lower():
                    if 'dcm' not in out: out['dcm'] = []
                    out['dcm'].append(os.path.join(opth, rfiles[i]['Name']))
                    
                elif '.bf' in rfiles[i]['Name'].lower():
                    if 'bf' not in out: out['bf'] = []
                    out['bf'].append(os.path.join(opth, rfiles[i]['Name']))
                    
                elif '.ima' in rfiles[i]['Name'].lower():
                    if 'ima' not in out: out['ima'] = []
                    out['ima'].append(os.path.join(opth, rfiles[i]['Name']))
                
    if len(rfiles)<1:
        print 'e> requested resources data is missing.'


    return out





#====================================================================================================
def download(
    xc,
    sbjix,
    fcomment='',
    tpoint=0,
    expmts=[],
    expid=[],
    download=False,
    prcl=False,
    t1w=False,
    t2w=False,
    umap=False,
    ute=False,
    pct=False,
    nrm=False,
    lm=False,
    cookie=''):

    '''
    expmts: list of dictionaries for XNAT experiments
    expid: [idnex for expmts or expt below when expmts is not given, time point label ordered according to the date]
    tpoint: time point according to date of scan, starting from 0 for baseline
    etc...
    '''

    # subject index in XNAT server
    # sbjix = xc['prjf']+ '_' + sbjno

    # get the experiment ID
    if not expmts:
        try:
            if cookie:
                expmts = get_xnatList(
                    xc['sbjs']+'/' +sbjix+ '/experiments?xsiType=xnat:petmrSessionData&format=json',
                    cookie=cookie)
            else:
                expmts = get_xnatList(
                    xc['sbjs']+'/' +sbjix+ '/experiments?xsiType=xnat:petmrSessionData&format=json',
                    usrpwd=xc['usrpwd'])

        except ValueError as ve:
            print '-------------------'
            print 'e> Value Error:', ve
            print '   for subject:', sbjix
            print '-------------------'
            raise

    #import pdb; pdb.set_trace()

    if expid:
        tpstr = 'TP'+str(expid[1])
    else:
        # dates for available time point scans
        dates = [datetime.strptime(d['date'], "%Y-%m-%d") for d in expmts]
        # date index for sorting
        didx = [k[0] for k in sorted(enumerate(dates), key=lambda x:x[1])]
        tpstr = 'TP'+str(tpoint)
        try:
            expid = [didx.index(np.int(tpoint)), tpoint]
        except ValueError:
            print 'e> time point {} does not exists!'.format(str(tpoint))
            return None

    # get the right experiment
    expt = expmts[expid[0]]
    if not expt:
        print 'w> no data for:', fcomment
        return None
        

    # core subject output path
    spth = os.path.join( xc['opth'], os.path.join(fcomment, tpstr ) )

    # norm path
    npth = os.path.join(spth, 'norm')
    # LM path
    lmpth = os.path.join(spth, 'LM')
    # T1/2 folder
    t1pth = os.path.join(spth, 'T1w')
    t2pth = os.path.join(spth, 'T2w')
    # pseudo-CT folder
    pctpth = os.path.join(spth, 'pCT')
    # DCM mu-map
    umpth = os.path.join(spth, 'umap')
    # UTE two sequences
    utepth = os.path.join(spth, 'UTE')
    # regional parcellations
    parpth = os.path.join(spth, 'prcl')

    

    # output dictionary
    out = {}

    # get all scans
    scans = get_xnatList(
        xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans', cookie=cookie
    )
    scntypes = [(s['type'],s['quality'],s['ID']) for s in scans]
    
    # import pdb; pdb.set_trace()

    # get the list of T1w files
    if t1w:
        out['T1nii'] = []
        if isinstance(t1w, basestring):
            pick_types = [s for s in scntypes if t1w == s[0]]
        if isinstance(t1w, list):
            pick_types = []
            for tt in t1w:
                pick_types += [s for s in scntypes if tt == s[0]]
        else:
            pick_types = [s for s in scntypes if 't1' in s[0].lower()]

        #-go through all scan types
        for sel_type in pick_types:
            TYPE = sel_type[0]
            QLTY = sel_type[1]
            ID = sel_type[2]
            t1files = get_xnatList(
                xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans/'+ID+'/resources/NIFTI/files',
                cookie=cookie
            )
            if download:
                opth = t1pth
                create_dir(opth)
                for i in range(len(t1files)):
                    # add 't1_' for the file to be recognised as T1w
                    fname = 't1_scan-'+ID+'_'+QLTY+'_'+TYPE+'_'+fcomment+'.'+t1files[i]['Name'].split('.',1)[-1]
                    status = get_xnatFile(
                        xc['url']+t1files[i]['URI'],
                        os.path.join(opth, fname),
                        cookie=cookie)
                    if status<0:
                        print 'e> no T1w image data:', fcomment
                    else:
                        out['T1nii'].append(os.path.join(opth, fname))
                if len(t1files)<1: 
                    print 'e> no T1w image data:', fcomment
            else:
                out['T1nii'] += len(t1files)
    # --------------------------------------------------------------------------------
    if t2w:
        out['T2nii'] = []
        if isinstance(t2w, basestring):
            pick_types = [s for s in scntypes if t2w == s[0]]
        if isinstance(t2w, list):
            pick_types = []
            for tt in t2w:
                pick_types += [s for s in scntypes if tt == s[0]]
        else:
            pick_types = [s for s in scntypes if 't1' in s[0].lower()]

        #-go through all scan types
        for sel_type in pick_types:
            TYPE = sel_type[0]
            QLTY = sel_type[1]
            ID = sel_type[2]
            files = get_xnatList(
                xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans/'+ID+'/resources/NIFTI/files',
                cookie=cookie
            )
            if download:
                opth = t2pth
                create_dir(opth)
                for i in range(len(files)):
                    # add 't2_' for the file to be recognised as T1w
                    fname = 't2_scan-'+ID+'_'+QLTY+'_'+TYPE+'_'+fcomment+'.'+files[i]['Name'].split('.',1)[-1]
                    status = get_xnatFile(
                        xc['url']+files[i]['URI'],
                        os.path.join(opth, fname),
                        cookie=cookie)
                    if status<0:
                        print 'e> no T1w image data:', fcomment
                    else:
                        out['T1nii'].append(os.path.join(opth, fname))
                if len(files)<1: 
                    print 'e> no T1w image data:', fcomment
            else:
                out['T2nii'] += len(files)
    # --------------------------------------------------------------------------------

    if ute:
        out['UTE1'], out['UTE2'] = [], []
        if isinstance(ute, basestring):
            pick_types = [s for s in scntypes if ute == s[0]]
        elif isinstance(ute, list):
            pick_types = []
            for tt in ute:
                pick_types += [s for s in scntypes if tt == s[0]]
        else:
            pick_types = [s for s in scntypes if 'ute' in s[0].lower()]

        print 'i> picked types:', pick_types

        for sel_type in pick_types:
            TYPE = sel_type[0]
            QLTY = sel_type[1]
            ID = sel_type[2]
            files = get_xnatList(
                xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans/'+ID+'/resources/DICOM/files',
                cookie=cookie
            )
            if download:
                create_dir(utepth)
                for i in range(len(files)):
                    status = get_xnatFile(
                        xc['url']+files[i]['URI'],
                        os.path.join(utepth, files[i]['Name']),
                        cookie=cookie
                    )
                    if status<0:
                        print 'e> no UTE data:', fcomment
                        raise IOError('Could not cownload the UTE data')
                        
                if len(files)<192 and [n['Name'] for n in files if '.zip' in n['Name']]:
                    fnm = [n['Name'] for n in files if '.zip' in n['Name']][0]
                    fzip = zipfile.ZipFile(os.path.join(utepth,fnm), 'r')
                    fzip.extractall(utepth)
                    fzip.close()
                # get the number of DICOM files
                out['#UTE'] = len([f for f in os.listdir(utepth) if f.endswith(dcm_ext)])
                # read one of the DICOM files to get the echo time
                fdcm = glob.glob(os.path.join(utepth, '*.dcm'))[0]
                dhdr = dcm.dcmread(fdcm)
                # get the Echo time of the UTE sequences
                if [0x018, 0x081] in dhdr:
                    ETstr = str(dhdr[0x018, 0x081].value)
                    # target path depending on the echo time
                    trgtpth = os.path.join(os.path.dirname(utepth), 'UTE_scan-'+ID+'_Q-'+QLTY+'_ET-'+ETstr.replace('.','-'))
                    os.rename(utepth,  trgtpth)
                else:
                    raise KeyError('Could  not find Echo Time in the DICOM header')
                if np.float32(ETstr)<0.1:
                    out['UTE1'] = trgtpth
                else:
                    out['UTE2'] = trgtpth

                        

    # DCM mu-map (UTE)
    if umap:
        um_TYPE = '1946_MRAC_UTE_UMAP'
        # get the list of norm files
        umfiles = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/scans/'+um_TYPE+'/resources/DICOM/files',
            cookie=cookie
        )
        # download the norm files:
        out['UTE'] = []
        if download:
            create_dir(umpth)
            for i in range(len(umfiles)):
                status = get_xnatFile(
                    xc['url']+umfiles[i]['URI'],
                    os.path.join(umpth, umfiles[i]['Name']),
                    cookie=cookie
                )
                if status<0:
                    print 'e> no UTE/Dixon mu-map data:', fcomment
                else:
                    out['UTE'].append(os.path.join(umpth, umfiles[i]['Name']))
            if len(umfiles)<192:
                if any(n['Name'] for n in umfiles if 'zip' in n['Name']):
                    fzip = zipfile.ZipFile(out['UTE'][0], 'r')
                    fzip.extractall(os.path.dirname(out['UTE'][0]))
                    fzip.close()
                    out['#UTE'] = len(os.listdir(os.path.dirname(out['UTE'][0])))
                else:
                    print 'e> UTE image data too small (<10):', fcomment
                    out['UTE'] = 'missing'
            elif len(umfiles)==192:
                out['#UTE'] = 192
        else:
            out['UTE'] = str(len(umfiles))

    #--------------------------------------------------------------------------------------
    # brain parcellations
    if prcl:
        PAR_TYPE = 'GIF_v2_TEST'
        parfiles = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/reconstructions/'+PAR_TYPE+'/files',
            cookie=cookie)
        if parfiles: out['prcl'] = []
        if download:
            create_dir(parpth)
            for i in range(len(parfiles)):
                status = get_xnatFile(
                    xc['url']+parfiles[i]['URI'], os.path.join(parpth, parfiles[i]['Name']),
                    cookie=cookie)
                if status<0:
                    print 'e> no parcellation image data:', fcomment
                else:
                    out['prcl'].append( os.path.join(parpth,parfiles[i]['Name']) )
            if len(parfiles)<5: 
                print 'e> it seems there is not enough parcellation data for:', fcomment
        else:
            out['prcl'] = str(len(parfiles))


    # pseudo-CT
    if pct:
        PCT_TYPE = 'PCT_HU'
        pctfiles = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/resources/'+PCT_TYPE+'/files',
            cookie=cookie)
        if download:
            create_dir(pctpth)
            for i in range(len(pctfiles)):
                status = get_xnatFile(
                    xc['url']+pctfiles[i]['URI'], os.path.join(pctpth, pctfiles[i]['Name']),
                    cookie=cookie)
                if status<0:
                    print 'e> no pCT image data:', fcomment
                else:
                    out['pCT'] = os.path.join(pctpth, pctfiles[i]['Name'])
            if len(pctfiles)<1: 
                print 'e> missing pCT file for', fcomment
        else:
            out['pCT'] = str(len(pctfiles))


    # get the list of norm files
    if nrm:
        nrmfiles = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/resources/Norm/files',
            cookie=cookie
        )
        # download the norm files:
        if download:
            create_dir(npth)
            for i in range(len(nrmfiles)):
                status = get_xnatFile(
                    xc['url']+nrmfiles[i]['URI'], os.path.join(npth, nrmfiles[i]['Name']),
                    cookie=cookie
                )
                if status<0:
                    if ip>=0: del plist[ip]
                    print 'e> no NORM data:', fcomment
                else:
                    if '.dcm' in nrmfiles[i]['Name'].lower():
                        out['nrm_dcm'] = os.path.join(npth, nrmfiles[i]['Name'])
                    elif '.bf' in nrmfiles[i]['Name'].lower():
                        out['nrm_bf'] = os.path.join(npth, nrmfiles[i]['Name'])
                    elif '.ima' in nrmfiles[i]['Name'].lower():
                        out['nrm_ima'] = os.path.join(npth, nrmfiles[i]['Name'])
            if len(nrmfiles)<1:
                print 'e> norm data missing.'
        else:
            out['norm'] = str(len(nrmfiles))

    # get the list of LM files
    if lm:
        lmfiles = get_xnatList(
            xc['sbjs']+'/' +sbjix+ '/experiments/' + expt['ID'] + '/resources/LM/files',
            cookie=cookie
        )
        # download the LM files:
        if download:
            create_dir(lmpth)
            for i in range(len(lmfiles)):
                # check if the file is already downloaded:
                if  os.path.isfile ( os.path.join(lmpth, lmfiles[i]['Name']) ) and \
                    str(os.path.getsize(os.path.join(lmpth, lmfiles[i]['Name'])))==lmfiles[i]['Size']:
                    print 'i> file of the same size,',lmfiles[i]['Name'], 'already exists: skipping download.'
                    if '.dcm' in lmfiles[i]['Name'].lower():
                        out['lm_dcm'] = os.path.join(lmpth, lmfiles[i]['Name'])
                    elif '.bf' in lmfiles[i]['Name'].lower():
                        out['lm_bf'] = os.path.join(lmpth, lmfiles[i]['Name'])
                    elif '.ima' in lmfiles[i]['Name'].lower():
                        out['lm_ima'] = os.path.join(lmpth, lmfiles[i]['Name'])
                else:
                    status = get_xnatFile(
                        xc['url']+lmfiles[i]['URI'], os.path.join(lmpth, lmfiles[i]['Name']),
                        cookie=cookie
                    )
                    if status<0:
                        print 'w> no LM data:', fcomment
                    else:
                        if '.dcm' in lmfiles[i]['Name'].lower():
                            out['lm_dcm'] = os.path.join(lmpth, lmfiles[i]['Name'])
                        elif '.bf' in lmfiles[i]['Name'].lower():
                            out['lm_bf'] = os.path.join(lmpth, lmfiles[i]['Name'])
                        elif '.ima' in lmfiles[i]['Name'].lower():
                            out['lm_ima'] = os.path.join(lmpth, lmfiles[i]['Name'])
            if len(lmfiles)<1:
                print 'e> LM data missing.'
        else:
            out['LM'] = str(len(lmfiles))

    out['folderin'] = spth

    return out













# #<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# def dobtp(xc, sbjid, sbjno, t0, t1, Nr, gpth, labels, dkeys, rois, txLUT, axLUT, Cnt):
#     # xc: dictionary of all properties for accessing XNAT server
#     # sbjid: subject ID
#     # sbjno: subject number
#     # t0: frame start time
#     # t1: frame stop time
#     # Nr: number of bootstrap realisations
#     # gpth: path to GIF parcellations
#     # labels: ROI labels based on the provided parcellation
#     # dkeys: keys (labels) used in the dictionary of the ROIs (easier/shorter for programming than the above labels)
#     # rois: the actual ROIs defined in numbers
#     # txLUT, axLUT, Cnt: transaxial, axial LUTs and scanner (Siemens mMR) constants

#     # subject index in XNAT server
#     sbjix = xc['prjf']+ '_' + sbjno

#     # core subject output path
#     spth = os.path.join(xc['opth'], sbjno+'_'+sbjid)
#     # processing output path
#     outpth = os.path.join(spth, 'output')
#     # LM path
#     lmpth = os.path.join(spth, 'LM')
    
#     # output file names and their xnat content tags
#     btpresults = [['SEimg_btp_'+str(Nr)+'.nii.gz',
#                    'Rimg_btp_' +str(Nr)+'.nii.gz',
#                    'bootstrap_boxplot_btp_'+str(Nr)+'.png'],
#                    ['SEBIMG','RBIMG','BTPBXP']]


#     #------------------------------------------------------
#     # check if the parcellation is available
#     g_lb = glob.glob( os.path.join(gpth, sbjid+'*labels.nii*') )
#     g_bc = glob.glob( os.path.join(gpth, sbjid+'*bc.nii*') ) 
#     if not (g_bc and g_lb):
#         print 'e> no GIF parcellation files available'
#         return {'pliststr':'m:par'}

#     # split the label file path
#     lbpth = os.path.split(g_lb[0])

#     # sum of all voxel values in any given ROI
#     roisum = {}
#     for k in dkeys: roisum[k] = [] 

#     # sum of the mask values <0,1> used for getting the mean value of any ROI
#     roimsk = {}
#     for k in dkeys: roimsk[k] = []
#     #------------------------------------------------------


#     #------------------------------------------------------
#     #GET THE RAW & INPUT DATA FOR RECONSTRUCTION
#     folderin = spth
#     datain = nipet.mmraux.explore_input(folderin, Cnt)
#     #------------------------------------------------------

#     #print the time difference between calibration and acquisition
#     nipet.mmraux.time_diff_norm_acq(datain)


#     # =============================================================
#     # R E C O N    W I T H    R O I    S A M P L I N G
#     # -------------------------------------------------------------
#     if Cnt['SPN']==11:
#         snshape = (Cnt['NSN11'], Cnt['NSANGLES'], Cnt['NSBINS'])
#     elif Cnt['SPN']==1:
#         snshape = (Cnt['NSN1'], Cnt['NSANGLES'], Cnt['NSBINS'])
#     print 'i> using this sino shape:', snshape

#     # for calculating variance online of scatter and random sinos
#     M1_ssn = np.zeros(snshape, dtype=np.float32)
#     M2_ssn = np.zeros(snshape, dtype=np.float32)

#     M1_sss = np.zeros(snshape, dtype=np.float32)
#     M2_sss = np.zeros(snshape, dtype=np.float32)

#     M1_rsn = np.zeros(snshape, dtype=np.float32)
#     M2_rsn = np.zeros(snshape, dtype=np.float32)

#     M1_img = np.zeros((Cnt['SO_IMZ'],Cnt['SO_IMY'],Cnt['SO_IMX']), dtype=np.float32)
#     M2_img = np.zeros((Cnt['SO_IMZ'],Cnt['SO_IMY'],Cnt['SO_IMX']), dtype=np.float32)

#     # sys.exit()


#     # do single non-bootstrap or multiple bootstrap (Nr) realisations
#     for itr in range((Nr-1)*(Cnt['BTP']>0)+1):

#         print ''
#         print '++++++++++++++++++++++++++++'
#         print 'i> performing realisation:', itr
#         print '++++++++++++++++++++++++++++'
#         print ''

#         # --------------------------------------------------------------
#         # histogram the LM data and save the results
#         hst = nipet.lm.mmrhist.hist(datain, txLUT, axLUT, Cnt, hst_store=False, t0=t0, t1=t1)

#         #get the mu-maps (hardware and object);  use the comment to save each affine transformation text file
#         cmmnt = '_btp'+str(Cnt['BTP'])+'_'+str(itr)
#         mumaps = nipet.img.mmrimg.get_mumaps(datain, Cnt, hst=hst, t0=t0, t1=t1, fcomment=cmmnt)

#         #update datain
#         datain = nipet.mmraux.explore_input(folderin, Cnt)

#         # do the proper recon with attenuation and scatter
#         recon = nipet.prj.mmrprj.osemone(datain, mumaps, hst, txLUT, axLUT, Cnt, recmod=3, itr=4, fwhm=0., store_img=True, ret_sct=True)
#         # name of saved file
#         fpetq = recon[1]
#         # reconstructed image
#         imcrr = recon[0]
#         # if bootstrap, calculate variance online
#         if Cnt['BTP']>0:
#             # calculate variance online:
#             nipet.mmr_auxe.varon(M1_ssn, M2_ssn, recon[3], itr, Cnt)
#             nipet.mmr_auxe.varon(M1_sss, M2_sss, recon[4], itr, Cnt)
#             nipet.mmr_auxe.varon(M1_rsn, M2_rsn, recon[6], itr, Cnt)
#             nipet.mmr_auxe.varon(M1_img, M2_img, imcrr, itr, Cnt)
#         # --------------------------------------------------------------


#         # --------------------------------------------------------------
#         # GIF: get the T1w and labels in register with PET
#         print '';  print ''
#         print 'i> extracting ROI values:'
#         # T1w:
#         t1dir = os.path.dirname(datain['T1nii'])
#         f_ = glob.glob( os.path.join(t1dir, '*_affine'+cmmnt+'.txt') )
#         if len(f_)>0:
#             frigid = f_[0]
#             print 'i> using rigid transform from this file:', frigid
#         else:
#             print 'e> no affine transformation text file found!'
#             return {'pliststr':'m:--'}

#         # PET (without scatter and attenuation corrections):
#         petdir = os.path.dirname(datain['em_nocrr'])
#         fpet = glob.glob( os.path.join(petdir, '*__ACbed.nii.gz') )
#         if len(fpet)==0:
#             print 'w> NAC PET image not found: using quantitative PET image for registration of MR to PET.'
#             fpet = fpetq
#         elif len(fpet)>1:
#             print 'w> multiple NAC/NSCT PET images found: using quantitative PET image for registration of MR to PET.'
#             fpet = fpetq
#         # resample the GIF T1/labels
#         #call the resampling routine to get the GIF T1/labels in PET space
#         if len(g_lb)>0 and len(fpet)>0:
#             fgt1 = os.path.join(t1dir, os.path.basename(g_lb[0]).split('.')[0]+ '_toPET' +'.nii.gz')
#             if os.path.isfile( Cnt['RESPATH'] ):
#                 call(
#                     [Cnt['RESPATH'],
#                     '-ref', fpet[0],
#                     '-flo', g_bc[0],
#                     '-trans', frigid,
#                     '-res', fgt1])
#             else:
#                 print 'e> path to resampling executable is incorrect!'
#                 sys.exit()

#         if len(g_lb)==0 or len(fpet)==0:
#             print 'e> no enough input data for resampling (e.g., floating labels or target PET image)!'
#             return {'pliststr':'m:--'}

#         # Get the labels before resampling to PET space
#         nilb = nib.load(g_lb[0])
#         A = nilb.get_sform()
#         imlb = nilb.get_data()
#         # get a copy of the image for masks (still in the original T1 orientation)
#         roi_img = np.copy(imlb)

#         for k in rois.keys():
#             roi_img[:] = 0
#             for i in rois[k]:
#                 roi_img[imlb==i] = 1.
#             # now save the image mask to nii file <froi1>
#             froi1 = os.path.join(t1dir, lbpth[1].split('.')[0][:8]+'_'+k+'.nii.gz')
#             nii_roi = nib.Nifti1Image(roi_img, A)
#             nii_roi.header['cal_max'] = 1.
#             nii_roi.header['cal_min'] = 0.
#             nib.save(nii_roi, froi1)
#             # file name for the resampled ROI to PET space
#             froi2 = os.path.join(t1dir, os.path.basename(g_lb[0]).split('.')[0]+ '_toPET_'+k+'.nii.gz')
#             if os.path.isfile( Cnt['RESPATH'] ):
#                 call(
#                     [Cnt['RESPATH'],
#                     '-ref', fpet[0],
#                     '-flo', froi1,
#                     '-trans', frigid,
#                     '-res', froi2])
#             else:
#                 print 'e> path to resampling executable is incorrect!'
#                 sys.exit()
#             # get the resampled ROI mask image
#             rmsk = nimpa.prc.getnii(froi2)
#             rmsk[rmsk>1.] = 1.
#             rmsk[rmsk<0.] = 0.

#             # erode the cerebral white matter region
#             if k=='EroWM':
#                 nilb = nib.load(froi2)
#                 B = nilb.get_sform()
#                 # tmp = nilb.get_data()
#                 tmp = ndi.filters.gaussian_filter(rmsk, nipet.mmraux.fwhm2sig(12./10), mode='mirror')
#                 rmsk = np.float32(tmp>0.85)
#                 froi3 = os.path.join(t1dir, os.path.basename(g_lb[0]).split('.')[0]+ '_toPET_'+k+'_eroded.nii.gz')
#                 nipet.img.mmrimg.savenii(rmsk, froi3, B, Cnt)

#             # ROI value and mask sums:
#             rvsum  = np.sum(imcrr*rmsk)
#             rmsum  = np.sum(rmsk)
#             roisum[k].append( rvsum )
#             roimsk[k].append( rmsum )
#             print ''
#             print '================================================================'
#             print 'i> ROI extracted:', k
#             print '   > value sum:', rvsum
#             print '   > mask sum :', rmsum
#             print '================================================================'
#         # --------------------------------------------------------------

#     # =============================================================

#     if Cnt['BTP']>0:
#         np.save(os.path.join(outpth,'saved_roisum_btp'+str(itr+1)+'.npy'), roisum)
#         np.save(os.path.join(outpth,'saved_roimsk_btp'+str(itr+1)+'.npy'), roimsk)
#         np.save(os.path.join(outpth,'saved_varon_btp'+str(itr+1)+'.npy'), (M1_rsn, M1_sss, M1_ssn, M2_rsn, M2_sss, M2_ssn, itr))
#         np.save(os.path.join(outpth,'saved_varon_img_btp'+str(itr+1)+'.npy'), (M1_img, M2_img, itr))
#     else:
#         np.save(os.path.join(outpth,'saved_roisum.npy'), roisum)
#         np.save(os.path.join(outpth,'saved_roimsk.npy'), roimsk)


#     # ANALYSIS
#     # get the ROIs for reference and neocortex regions
#     ROIs = np.zeros((len(dkeys),itr+1), dtype=np.float32)
#     for i in range(len(dkeys)):
#         ROIs[i,:] = np.array(roisum[dkeys[i]]) / np.array(roimsk[dkeys[i]])

#     # pick the reference regions for reporting
#     refi = [0,2,3]
#     # pick the neocortex regions for reporting (boxplot)
#     roii = range(7, len(dkeys))
#     roilables = [labels[k] for k in roii]
#     print 'i> chosen regions of interests:',[labels[i] for i in refi]
#     print 'i> chosen regions of interests:', roilables


#     # one plot per ROI
#     if Cnt['BTP']>0:
#         ip = 1
#         plt.figure(figsize=(16, 12))
#         for i in range(len(refi)):
#             print 'i> ref region:', labels[refi[i]]
#             roi = np.array(roisum[dkeys[refi[i]]]) / np.array(roimsk[dkeys[refi[i]]])
#             roi.shape = (1, itr+1)
#             roi = np.repeat(roi, len(roii), axis=0)
#             suvr = ROIs[roii,:]/roi
#             plt.subplot(1,3,ip)
#             if mplv>15:       plt.boxplot(suvr.T, showmeans=True, labels=roilables)
#             else:             plt.boxplot(suvr.T, labels=roilables)
#             plt.axhline(y=suvr[-1,:].mean(),linewidth=1, color='b')
#             plt.xticks(range(1,1+len(roii)), rotation='vertical')
#             plt.title(labels[refi[i]]+' as reference')
#             plt.subplots_adjust(bottom=0.20)
#             ip += 1
#         plt.savefig(os.path.join(outpth, btpresults[0][2]), format='png', dpi=300)

#         # ratio (aka. SUVR) and SE image
#         Rimg = M1_img/np.mean(ROIs[0,:])
#         SEimg = ( (M2_img/itr)**.5 ) / np.mean(ROIs[0,:])
#         nipet.img.mmrimg.savenii(SEimg, os.path.join(outpth, btpresults[0][0]), Cnt['AFFINE'], Cnt)
#         nipet.img.mmrimg.savenii(Rimg,  os.path.join(outpth, btpresults[0][1]), Cnt['AFFINE'], Cnt)

#         # write ROIs to csv file
#         np.savetxt( os.path.join(outpth,'bootstrap_ROIsum.csv'), ROIs.T, delimiter=',', header=','.join(fieldnames) )


#     # save the boxplot and the NIfTI images to XNAT
#     if xc['upload']:
#         xnatURL = xc['sbjs']+'/' +sbjix+ '/experiments/' + expt[0]['ID'] + '/resources/BTP'
#         nipet.xnat.qc_xnat.put_xnat(xc['usrpwd'], xnatURL+'?xsi:type=xnat:resourceCatalog&format=PNG')
#         nipet.xnat.qc_xnat.put_xnatFile(xc['usrpwd'], xnatURL+'/files?content='+btpresults[1][2], os.path.join(outpth, btpresults[0][2]) )

#         nipet.xnat.qc_xnat.put_xnat(xc['usrpwd'], xnatURL+'?xsi:type=xnat:resourceCatalog&format=NIFTI')
#         nipet.xnat.qc_xnat.put_xnatFile(xc['usrpwd'], xnatURL+'/files?content='+btpresults[1][1], os.path.join(outpth, btpresults[0][1]) )
#         nipet.xnat.qc_xnat.put_xnatFile(xc['usrpwd'], xnatURL+'/files?content='+btpresults[1][0], os.path.join(outpth, btpresults[0][0]) )

#     dic_out = { 'pliststr':'y',
#                 'roisum':roisum, 'roimsk':roimsk, 'M1_rsn':M1_rsn, 'M1_sss':M1_sss, 'M1_ssn':M1_ssn, 'M2_rsn':M2_rsn,
#                 'M2_sss':M2_sss, 'M2_ssn':M2_ssn, 'M1_img':M1_img, 'M2_img':M1_img, 'itr':itr, 'ROIs':ROIs, 'labels':labels }
    
#     #remove the downloaded LM file
#     if xc['rmvlm']:   os.remove(os.path.join(lmpth, lmfiles[i]['Name']))

#     return dic_out





"""auxilary imaging functions for PET image reconstruction and analysis."""

__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"


import numpy as np
import os
from niftypet import nimpa

#==============================================================================

def obtain_image(img, Cnt=[], imtype='', verbose=False):
    ''' 
    Obtain the image (hardware or object mu-map) from file,
    numpy array, dictionary or empty list (assuming blank then).
    The image has to have the dimensions of the PET image used as in Cnt['SO_IM[X-Z]'].
    '''

    if Cnt: verbose = Cnt['VERBOSE']
    # establishing what and if the image object has been provided
    # all findings go to the output dictionary
    output = {}
    if isinstance(img, dict):
        if Cnt and img['im'].shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            print 'e> provided '+imtype+' via the dictionary has inconsistent dimensions compared to Cnt.'
            raise ValueError('Wrong dimensions of the mu-map')
        else:
            output['im'] = img['im']
            output['exists'] = True
            output['fim'] = img['fim']
            if 'faff' in img: output['faff'] = img['faff']
            if 'fmuref' in img: output['fmuref'] = img['fmuref']
            if 'affine' in img: output['affine'] = img['affine']
            if verbose: print 'i> using '+imtype+' from dictionary.'

    elif isinstance(img, (np.ndarray, np.generic) ):
        if Cnt and img.shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
            print 'e> provided '+imtype+' via the numpy array has inconsistent dimensions compared to Cnt.'
            raise ValueError('Wrong dimensions of the mu-map')
        else:
            output['im'] = img
            output['exists'] = True
            output['fim'] = ''
            if verbose: print 'i> using hardware mu-map from numpy array.'

    elif isinstance(img, basestring):
        if os.path.isfile(img):
            imdct = nimpa.getnii(img, output='all')
            output['im'] = imdct['im']
            output['affine'] = imdct['affine']
            if Cnt and output['im'].shape!=(Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']):
                print 'e> provided '+imtype+' via file has inconsistent dimensions compared to Cnt.'
                raise ValueError('Wrong dimensions of the mu-map')
            else:
                output['exists'] = True
                output['fim'] = img
                if verbose: print 'i> using '+imtype+' from NIfTI file.'
        else:
            print 'e> provided '+imtype+' path is invalid.'
            return None
    elif isinstance(img, list):
        output['im'] = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
        if verbose: print 'w> '+imtype+' has not been provided -> using blank.'
        output['fim'] = ''
        output['exists'] = False
    #------------------------------------------------------------------------
    return output


#==============================================================================

def dynamic_timings(flist, offset=0):
    '''
    Get start and end frame timings from a list of dynamic PET frame definitions.
    flist can be 1D list of time duration for each dynamic frame, e.g.: flist = [15, 15, 15, 15, 30, 30, 30, ...]
    or a 2D list of lists having 2 entries: first for the number of repetitions and the other for the frame duration,
    e.g.: flist = [[4,15], [3,15], ...].
    offset adjusts for the start time (usually when prompts are strong enough over randoms)
    The output is a dictionary:
    out['timings'] = [[0, 15], [15, 30], [30, 45], [45, 60], [60, 90], [90, 120], [120, 150], ...]
    out['total'] = total time
    out['frames'] = array([ 15,  15,  15,  15,  30,  30,  30,  30, ...])

    '''
    if not isinstance(flist, list):
        raise TypeError('Wrong type of frame data input')
    if all([isinstance(t,(int, np.int32, np.int16, np.int8, np.uint8, np.uint16, np.uint32)) for t in flist]):
        tsum = offset
        # list of frame timings
        if offset>0:
            t_frames = [[0, offset]]
        else:
            t_frames = []
        for i in range(len(flist)):
            # frame start time
            t0 = tsum
            tsum += flist[i]
            # frame end time
            t1 = tsum
            # append the timings to the list
            t_frames.append([t0, t1])
        frms = np.uint16(flist)
    elif all([isinstance(t,list) and len(t)==2 for t in flist]):
        if offset>0:
            flist.insert(0,[1,offset])
            farray = np.asarray(flist, dtype=np.uint16)
        else:
            farray = np.array(flist)
        # number of dynamic frames
        nfrm = np.sum(farray[:,0])
        # list of frame duration
        frms = np.zeros(nfrm,dtype=np.uint16)
        #frame iterator
        fi = 0
        #time sum of frames
        tsum = 0
        # list of frame timings
        t_frames = []
        for i in range(0, farray.shape[0]):
            for t in range(0, farray[i,0]):
                # frame start time
                t0 = tsum
                tsum += farray[i,1]
                # frame end time
                t1 = tsum
                # append the timings to the list
                t_frames.append([t0, t1])
                frms[fi] = farray[i,1]
                fi += 1
    else:
        raise TypeError('Unrecognised data input.')
    # prepare the output dictionary
    out = {'total':tsum, 'frames':frms, 'timings':t_frames}
    return out
#=================================================================================================